//-----------------------------------------------------------------------------
// The confidential and proprietary information contained in this file may
// only be used by a person authorised under and to the extent permitted
// by a subsisting licensing agreement from ARM Limited or its affiliates.
//
//            (C) COPYRIGHT 2012-2020 ARM Limited or its affiliates.
//                ALL RIGHTS RESERVED
//
// This entire notice must be reproduced on all copies of this file
// and copies of this file may only be made by a person if such person is
// permitted to do so under the terms of a subsisting license agreement
// from ARM Limited or its affiliates.
//
//            Release Information : CORTEX-A65AE-MP083-r1p0-00eac0
//
//-----------------------------------------------------------------------------
// SystemVerilog-2012 (IEEE Std 1800-2012)
//-----------------------------------------------------------------------------



module helios_dpu_crypto_pmull64 (
  input  logic  [63:0] opa_i,
  input  logic  [63:0] opb_i,
  output logic [126:0] pmull64_res_o
);


  function automatic logic [6:0] pmul_4x4(input logic [3:0] op_a, input logic [3:0] op_b);

    logic [6:0] result;

    result = '0;
    for(integer i=0; i<4; i++) begin
      result[i+:4] ^= ({4{op_a[i]}} & op_b);
    end

    pmul_4x4 = result;

  endfunction


  function automatic logic [14:0] pmul_8x8(input logic [7:0] op_a, input logic [7:0] op_b);

    logic [3:0] kara_op_a; 
    logic [3:0] kara_op_b; 
    logic [6:0] pp_lo;     
    logic [6:0] pp_kar;    
    logic [6:0] pp_hi;     

    kara_op_a = op_a[7:4] ^ op_a[3:0];
    kara_op_b = op_b[7:4] ^ op_b[3:0];

    pp_lo  = pmul_4x4(.op_a(op_a[3:0]), .op_b(op_b[3:0]));
    pp_kar = pmul_4x4(.op_a(kara_op_a), .op_b(kara_op_b));
    pp_hi  = pmul_4x4(.op_a(op_a[7:4]), .op_b(op_b[7:4]));

    pmul_8x8 = {pp_hi, {8{1'b0}}} ^ {{4{1'b0}}, pp_hi, {4{1'b0}}} ^ {{4{1'b0}}, pp_kar, {4{1'b0}}} ^ {{4{1'b0}}, pp_lo, {4{1'b0}}} ^ {{8{1'b0}}, pp_lo};

  endfunction


  function automatic logic [30:0] pmul_16x16(input logic [15:0] op_a, input logic [15:0] op_b);

    logic  [7:0] kara_op_a; 
    logic  [7:0] kara_op_b; 
    logic [14:0] pp_lo;     
    logic [14:0] pp_kar;    
    logic [14:0] pp_hi;     

    kara_op_a = op_a[15:8] ^ op_a[7:0];
    kara_op_b = op_b[15:8] ^ op_b[7:0];

    pp_lo  = pmul_8x8(.op_a(op_a[7:0]),  .op_b(op_b[7:0]));
    pp_kar = pmul_8x8(.op_a(kara_op_a),  .op_b(kara_op_b));
    pp_hi  = pmul_8x8(.op_a(op_a[15:8]), .op_b(op_b[15:8]));

    pmul_16x16 = {pp_hi, {16{1'b0}}} ^ {{8{1'b0}}, pp_hi, {8{1'b0}}} ^ {{8{1'b0}}, pp_kar, {8{1'b0}}} ^ {{8{1'b0}}, pp_lo, {8{1'b0}}} ^ {{16{1'b0}}, pp_lo};

  endfunction


  function automatic logic [62:0] pmul_32x32(input logic [31:0] op_a, input logic [31:0] op_b);

    logic [15:0] kara_op_a; 
    logic [15:0] kara_op_b; 
    logic [30:0] pp_lo;     
    logic [30:0] pp_kar;    
    logic [30:0] pp_hi;     

    kara_op_a = op_a[31:16] ^ op_a[15:0];
    kara_op_b = op_b[31:16] ^ op_b[15:0];

    pp_lo  = pmul_16x16(.op_a(op_a[15:0]),  .op_b(op_b[15:0]));
    pp_kar = pmul_16x16(.op_a(kara_op_a),   .op_b(kara_op_b));
    pp_hi  = pmul_16x16(.op_a(op_a[31:16]), .op_b(op_b[31:16]));

    pmul_32x32 = {pp_hi, {32{1'b0}}} ^ {{16{1'b0}}, pp_hi, {16{1'b0}}} ^ {{16{1'b0}}, pp_kar, {16{1'b0}}} ^ {{16{1'b0}}, pp_lo, {16{1'b0}}} ^ {{32{1'b0}}, pp_lo};

  endfunction



  logic [31:0] kara_op_a; 
  logic [31:0] kara_op_b; 
  logic [62:0] pp_lo;     
  logic [62:0] pp_kar;    
  logic [62:0] pp_hi;     



  assign kara_op_a = opa_i[63:32] ^ opa_i[31:0];
  assign kara_op_b = opb_i[63:32] ^ opb_i[31:0];

  assign pp_lo  = pmul_32x32(.op_a(opa_i[31:0]),  .op_b(opb_i[31:0]));
  assign pp_kar = pmul_32x32(.op_a(kara_op_a),    .op_b(kara_op_b));
  assign pp_hi  = pmul_32x32(.op_a(opa_i[63:32]), .op_b(opb_i[63:32]));

  assign pmull64_res_o = {pp_hi, {64{1'b0}}} ^ {{32{1'b0}}, pp_hi, {32{1'b0}}} ^ {{32{1'b0}}, pp_kar, {32{1'b0}}} ^ {{32{1'b0}}, pp_lo, {32{1'b0}}} ^ {{64{1'b0}}, pp_lo};

endmodule

